// SPDX-License-Identifier: MIT
/*
 * Copyright (c) 2023 Atmark Techno,Inc.
 */

#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>

#include "device-info.h"

char *version = "1.1-r0";
bool is_print_env = false;

/*
 * Long options without short options
 * are defined by enums starting with CHAR_MAX + 1
 */
enum {
	SE_PARAM_OPTION = CHAR_MAX + 1
};

static struct option long_options[] = {
	{"abos-version",	no_argument,	NULL,	'a'},
	{"additional-board",	no_argument,	NULL,	'b'},
	{"env",	no_argument,	NULL,	'e'},
	{"help",	no_argument,	NULL,	'h'},
	{"mac-address",	no_argument,	NULL,	'm'},
	{"product",	no_argument,	NULL,	'p'},
	{"serial",	no_argument,	NULL,	's'},
	{"version",	no_argument,	NULL,	'v'},
	{"se-param",	no_argument,	NULL,	SE_PARAM_OPTION},
	{0,			0,				0,	0}
};

static void __attribute__ ((__noreturn__)) print_version(char *argv0) {
	printf("%s version %s\n", argv0, version);
	exit(0);
}

static void __attribute__ ((__noreturn__)) help(char *argv0) {
	printf("Usage: %s [options]\n", argv0);
	printf("Options:\n");
	printf("  -a, --abos-version: show ABOS version\n");
	printf("  -b, --additional-board: show additional board info\n");
	printf("  -e, --env: show as environmental variables\n");
	printf("  -h, --help: show this help\n");
	printf("  -m, --mac-address: print the MAC address of this board\n");
	printf("  -p, --product: print the product name of this board\n");
	printf("  -s, --serial: print the serial number of this board\n");
	printf("  -v, --version: show device-info version\n");
	printf("  --se-param: print param of secure element in a format that\n");
	printf("              can be exported directry to EX_SSS_BOOT_SSS_PORT\n");
	exit(0);
}

int main(int argc, char *argv[])
{
	int c;
	int option_index = 0;
	int ret = 0;

	bool is_print_abos_version = false;
	bool is_print_mac_addr = false;
	bool is_print_product_name = false;
	bool is_print_serial_num = false;
	bool is_print_se_param = false;
	bool is_print_ext_board_info = false;

	while ((c = getopt_long(argc, argv, "abehmpsv", long_options, &option_index)) > 0) {

		switch (c) {
		case 'a':
			is_print_abos_version = true;
			break;
		case 'b':
			is_print_ext_board_info = true;
			break;
		case 'e':
			is_print_env = true;
			break;
		case 'h':
			help(argv[0]);
		case 'm':
			is_print_mac_addr = true;
			break;
		case 'p':
			is_print_product_name = true;
			break;
		case 's':
			is_print_serial_num = true;
			break;
		case 'v':
			print_version(argv[0]);
		case SE_PARAM_OPTION:
			is_print_se_param = true;
			break;
		default:
			fprintf(stderr, "Execute '%s --help'.\n", argv[0]);
			return 1;
		}
	}
	if (argc == 1) {
    	help(argv[0]);
	}

	if (optind < argc) {
		printf("Unrecognized option: ");
		while (optind < argc)
			printf("%s ", argv[optind++]);
		printf("\n");
	}

	if (is_print_abos_version)
		ret |= print_abos_version();
	if (is_print_mac_addr)
		ret |= print_mac_addr();
	if (is_print_product_name)
		ret |= print_product_name();
	if (is_print_serial_num)
		ret |= print_serial_num();
	if (is_print_se_param)
		ret |= print_se_param();
	if (is_print_ext_board_info)
		ret |= print_ext_board_info();

	return ret;
}
